



from .Console import *
from .ConsoleBufferRGB_Mixins import *
from ._Rect import _Rect










_ConsoleBufferRGBCell = typing.NewType("_ConsoleBufferRGBCell", object)

class _ConsoleBufferRGBCell(object):

	def __init__(self, colorBG:int, colorFG:int):
		self._lastColBG = colorBG
		self._lastColFG = colorFG
		self._lastC = " "
		self.colBG = colorBG
		self.colFG = colorFG
		self.c = " "
		self.bChanged = False
	#

	def copyTo(self, other:_ConsoleBufferRGBCell):
		other.colBG = self.colBG
		other.colFG = self.colFG
		other.c = self.c
		other.bChanged = True
	#

#



#
# This class implements a buffer for the text console. It consists of individual cells, one for every character.
#
# Every cell contain the following attributes:
# * `str c` : The (single) character of this cell.
# * `bool bChanged` : A boolean flag set to `true` on change. (Typically only changed cells will be rendered.)
# * `int bgColor` : The background color of the cell.
#
class ConsoleBufferRGB(ConsoleBufferRGB_Mixins):

	################################################################################################################################
	## Constants
	################################################################################################################################

	################################################################################################################################
	## Constructor
	################################################################################################################################

	#
	# Constructor method.
	#
	def __init__(self, width:int = None, height:int = None, bgColor:int = None, bWithCaching:bool = True):
		self._bgColor = bgColor
		self._width = Console.width() - 1 if width is None else width
		self._height = Console.height() - 1 if height is None else height
		self._widthM1 = self._width - 1
		self._heightM1 = self._height - 1
		self._numberOfCells = self._width * self._height

		self._lineNumbers = tuple(range(0, self._height))
		self._columnNumbers = tuple(range(0, self._width))

		self._data = [ [_ConsoleBufferRGBCell(bgColor, None) for x in self._columnNumbers] for y in self._lineNumbers]

		assert len(self._lineNumbers) == self._height
		assert len(self._columnNumbers) == self._width
		assert len(self._data) == self._height
		for i in range(0, self._height):
			assert len(self._data[i]) == self._width

		self._bFullPaint = True

		# caching

		self._bWithCaching = bWithCaching
		if bWithCaching:
			self.__colMapBG = {}
			self.__colMapFG = {}
			self.__compileColor = self.__compileColorWithCaching
		else:
			self.__colMapBG = None
			self.__colMapFG = None
			self.__compileColor = self.__compileColorWithoutCaching

		#self._Mixins_init()

		self.defaultFramedBoxSettings = FramedBoxSettingsRGB()
	#

	################################################################################################################################
	## Public Properties
	################################################################################################################################

	#
	# Returns a tuple with two values: The width and height.
	#
	@property
	def size(self):
		return self._width, self._height
	#

	@property
	def width(self):
		return self._width
	#

	@property
	def height(self):
		return self._height
	#

	################################################################################################################################
	## Helper Methods
	################################################################################################################################

	#
	# Convert the rgb color values to a string
	#
	def __compileColorWithCaching(self, bgRGB:int, fgRGB:int):
		s = ""
		if bgRGB != None:
			colStr = self.__colMapBG.get(bgRGB)
			if colStr is None:
				colStr = "\033[48;2;" + str((bgRGB >> 16) & 0xff) + ";" + str((bgRGB >> 8) & 0xff) + ";" + str(bgRGB & 0xff) + "m"
				self.__colMapBG[bgRGB] = colStr
			s += colStr
		if fgRGB != None:
			colStr = self.__colMapFG.get(fgRGB)
			if colStr is None:
				colStr = "\033[38;2;" + str((fgRGB >> 16) & 0xff) + ";" + str((fgRGB >> 8) & 0xff) + ";" + str(fgRGB & 0xff) + "m"
				self.__colMapFG[fgRGB] = colStr
			s += colStr
		return s
	#

	#
	# Convert the rgb color values to a string
	#
	def __compileColorWithoutCaching(self, bgRGB:int, fgRGB:int):
		s = ""
		if bgRGB != None:
			s += "\033[48;2;" + str((bgRGB >> 16) & 0xff) + ";" + str((bgRGB >> 8) & 0xff) + ";" + str(bgRGB & 0xff) + "m"
		if fgRGB != None:
			s += "\033[38;2;" + str((fgRGB >> 16) & 0xff) + ";" + str((fgRGB >> 8) & 0xff) + ";" + str(fgRGB & 0xff) + "m"
		return s
	#

	################################################################################################################################
	## Public Methods
	################################################################################################################################

	def clearInternalCache(self):
		if self._bWithCaching:
			self.__colMapBG.clear()
			self.__colMapFG.clear()
	#

	def __str__(self):
		return "<ConsoleBufferRGB size=%dx%d at %x>" % (self._width, self._height, id(self))
	#

	def __repr__(self):
		return "<ConsoleBufferRGB size=%dx%d at %x>" % (self._width, self._height, id(self))
	#

	#
	# Force full repaint for next rendering.
	#
	def fullRepaint(self):
		self._bFullPaint = True
	#

	def bufferToBuffer(self, ofsX:int, ofsY:int, other, bForceFullRepaint:bool = False):
		assert isinstance(other, ConsoleBufferRGB)

		sourceRect = _Rect(0, 0, self._width, self._height)
		destRect = _Rect(0, 0, other._width, other._height)
		destRect0 = sourceRect.clone().move(ofsX, ofsY).intersect(destRect)
		sourceRect0 = destRect0.clone().move(-ofsX, -ofsY)

		for y in range(0, destRect0.height):
			currentRowSrc = self._data[sourceRect0.y + y]
			currentRowDest = other._data[destRect0.y + y]
			for x in range(0, destRect0.width):
				currentCellSrc = currentRowSrc[sourceRect0.x + x]
				currentCellDest = currentRowDest[destRect0.x + x]
				currentCellDest.c = currentCellSrc.c
				currentCellDest.colBG = currentCellSrc.colBG
				currentCellDest.colFG = currentCellSrc.colFG
				currentCellDest.bChanged = True
	#

	def bufferToConsole(self, ofsX:int = 0, ofsY:int = 0, bForceFullRepaint:bool = False):
		if self._bFullPaint or bForceFullRepaint:
			characters = []
			for y in self._lineNumbers:
				characters.clear()
				lastColBG = None
				lastColFG = None

				currentRow = self._data[y]
				for x in self._columnNumbers:
					currentCell = currentRow[x]
					#assert currentCell.col != None
					#assert currentCell.c != None
					currentCell._lastColBG = currentCell.colBG
					currentCell._lastColFG = currentCell.colFG
					currentCell._lastC = currentCell.c
					if (lastColBG != currentCell.colBG) or (lastColFG != currentCell.colFG):
						lastColBG = currentCell.colBG
						lastColFG = currentCell.colFG
						characters.append(self.__compileColor(lastColBG, lastColFG))
					characters.append(currentCell.c)
					currentCell.bChanged = False
				characters.append(Console.RESET)
				Console.printAt(ofsX, ofsY + y, "".join(characters), False)
			self._bFullPaint = False

		else:
			startX = -1
			characters = []
			bLastWasChanged = False
			lastColBG = None
			lastColFG = None

			for y in self._lineNumbers:
				currentRow = self._data[y]
				for x in self._columnNumbers:
					currentCell = currentRow[x]
					#assert currentCell.col != None
					#assert currentCell.c != None
					if currentCell.bChanged:
						# we might have a change here
						if (currentCell.c != currentCell._lastC) \
							or (currentCell.colBG != currentCell._lastColBG) \
							or (currentCell.colFG != currentCell._lastColFG):

							# yes, we really have a change here
							currentCell._lastColBG = currentCell.colBG
							currentCell._lastColFG = currentCell.colFG
							currentCell._lastC = currentCell.c
							if bLastWasChanged:
								if (lastColBG != currentCell.colBG) or (lastColFG != currentCell.colFG):
									lastColBG = currentCell.colBG
									lastColFG = currentCell.colFG
									characters.append(self.__compileColor(lastColBG, lastColFG))
								characters.append(currentCell.c)
							else:
								if startX >= 0:
									characters.append(Console.RESET)
									Console.printAt(ofsX + startX, ofsY + y, "".join(characters), False)
									characters.clear()
								startX = x
								lastColBG = currentCell.colBG
								lastColFG = currentCell.colFG
								characters.append(self.__compileColor(currentCell.colBG, currentCell.colFG))
								characters.append(currentCell.c)
								bLastWasChanged = True
						else:
							bLastWasChanged = False
						currentCell.bChanged = False
					else:
						bLastWasChanged = False

				if startX >= 0:
					characters.append(Console.RESET)
					Console.printAt(ofsX + startX, ofsY + y, "".join(characters), False)
					characters.clear()
					startX = -1
					bLastWasChanged = False
					lastColBG = None
					lastColFG = None

		print(Console.RESET_TOPLEFT, end="", flush=True)
	#

	################################################################################################################################
	## Public Static Methods
	################################################################################################################################

#








import re





class ViewPortRGB(object):

	def __init__(self, ri, gi, bi):
		self.ri = ri
		self.gi = gi
		self.bi = bi
	#

	def __int__(self):
		return self.ri * 256 + self.gi * 16 + self.bi
	#

	def __str__(self):
		return "#%02x%02x%02x" % (self.ri * 16, self.gi * 16, self.bi * 16)
	#

	_HEXVALS = "0123456789abcdef"
	_P1 = re.compile(r"^#([0-9a-f][0-9a-f])([0-9a-f][0-9a-f])([0-9a-f][0-9a-f])$")
	_P2 = re.compile(r"^#([0-9a-f])([0-9a-f])([0-9a-f])$")
	_P3 = re.compile(r"^rgb\(\s*([0-9]+)\s*,\s*([0-9]+)\s*,\s*([0-9]+)\s*\)$")

	#
	# Parse CSS codes as used by web technologies
	#
	@staticmethod
	def parseCSS(css:str):
		css = css.lower()
		m = ViewPortRGB._P1.match(css)
		if m:
			sr = m.group(1)
			sg = m.group(2)
			sb = m.group(3)
			r = (ViewPortRGB._HEXVALS.index(sr[0]) << 4) + ViewPortRGB._HEXVALS.index(sr[1])
			g = (ViewPortRGB._HEXVALS.index(sg[0]) << 4) + ViewPortRGB._HEXVALS.index(sg[1])
			b = (ViewPortRGB._HEXVALS.index(sb[0]) << 4) + ViewPortRGB._HEXVALS.index(sb[1])
			return ViewPortRGB(r >> 4, g >> 4, b >> 4)
		m = ViewPortRGB._P2.match(css)
		if m:
			sr = m.group(1)
			sg = m.group(2)
			sb = m.group(3)
			r = ViewPortRGB._HEXVALS.index(sr)
			g = ViewPortRGB._HEXVALS.index(sg)
			b = ViewPortRGB._HEXVALS.index(sb)
			return ViewPortRGB(r, g, b)
		m = ViewPortRGB._P3.match(css)
		if m:
			sr = m.group(1)
			sg = m.group(2)
			sb = m.group(3)
			r = int(sr)
			if r > 255:
				raise Exception("Invalid red component in: " + repr(css))
			g = int(sg)
			if g > 255:
				raise Exception("Invalid green component in: " + repr(css))
			b = int(sb)
			if b > 255:
				raise Exception("Invalid blue component in: " + repr(css))
			return ViewPortRGB(r >> 4, g >> 4, b >> 4)
		raise Exception("Not a CSS color string: " + repr(css))
	#

	@staticmethod
	def rgb256(r:int, g:int, b:int):
		if (r < 0) or (r > 255):
			raise Exception("Red value must be a valid integer value! (Value specified: " + str(r) + ")")
		if (g < 0) or (g > 255):
			raise Exception("Red value must be a valid integer value! (Value specified: " + str(g) + ")")
		if (b < 0) or (b > 255):
			raise Exception("Red value must be a valid integer value! (Value specified: " + str(b) + ")")
		return ViewPortRGB(r >> 4, g >> 4, b >> 4)
	#

	@staticmethod
	def rgb1(r:float, g:float, b:float):
		if not isinstance(g, (int, float)) or (r < 0) or (r > 1):
			raise Exception("Red value must be a valid float value in the range [0..1]! (Value specified: " + str(r) + ")")
		if not isinstance(g, (int, float)) or (g < 0) or (g > 1):
			raise Exception("Green value must be a valid float value in the range [0..1]! (Value specified: " + str(g) + ")")
		if not isinstance(g, (int, float)) or (b < 0) or (b > 1):
			raise Exception("Blue value must be a valid float value in the range [0..1]! (Value specified: " + str(b) + ")")
		return ViewPortRGB(int(r * 15.9999), int(g * 15.9999), int(b * 15.9999))
	#

#









import sys
import tty
import termios
import typing
from select import select




def readchar():
	fd = sys.stdin.fileno()
	old_settings = termios.tcgetattr(fd)
	try:
		tty.setraw(sys.stdin.fileno())
		ch = sys.stdin.read(1)
	finally:
		termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
	return ch
#

def readchar_loop():
	fd = sys.stdin.fileno()
	old_settings = termios.tcgetattr(fd)
	try:
		tty.setraw(sys.stdin.fileno())
		while True:
			yield sys.stdin.read(1)
	finally:
		termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
#

def readkeydata_loop(timeout:typing.Union[float,int,None] = 0):
	if timeout is None:
		pass
	else:
		assert isinstance(timeout, (float,int))

	fd = sys.stdin.fileno()
	old_settings = termios.tcgetattr(fd)
	try:
		tty.setraw(fd)
		mode = termios.tcgetattr(fd)
		mode[6][termios.VMIN] = 0
		mode[6][termios.VTIME] = 0
		termios.tcsetattr(fd, termios.TCSAFLUSH, mode)

		if timeout:
			while True:
				rlist, _, _ = select([fd], [], [], timeout)
				if rlist:
					yield sys.stdin.buffer.read(4096)
				else:
					yield None

		else:
			while True:
				yield sys.stdin.buffer.read(4096)

	finally:
		termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
#



















from .IntRGB import IntRGB




class FramedBoxSettingsRGB(object):

	def __init__(self):
		self.bgColor = IntRGB.parseCSS("#000000")
		self.bgChar = " "
		self.frameColorBG = IntRGB.parseCSS("#000000")
		self.frameColorFG = IntRGB.parseCSS("#e0e0e0")
		self.frameCharLeft = "|"
		self.frameCharRight = "|"
		self.frameCharTop = "-"
		self.frameCharBottom = "-"
		self.frameCharTopLeft = "/"
		self.frameCharTopRight = "\\"
		self.frameCharBottomLeft = "\\"
		self.frameCharBottomRight = "/"
		self.textColorBG = IntRGB.parseCSS("#000000")
		self.textColorFG = IntRGB.parseCSS("#e0e0e0")
	#

#












from .Console import *
from .ConsoleBufferRGB_Mixins import *
from ._Rect import _Rect
from .RGBColorMapper import RGBColorMapper
from .FramedBoxSettingsRGB import FramedBoxSettingsRGB
from .SimpleTable import *







_ConsoleBufferRGBCell2 = typing.NewType("_ConsoleBufferRGBCell2", object)

class _ConsoleBufferRGBCell2(object):

	def __init__(self, colorBG:int, colorFG:int, c:str = " "):
		self.data = [ colorBG, colorFG, c ]
	#

	def copyTo(self, other:_ConsoleBufferRGBCell2):
		if other.data != self.data:
			# other.data = list(self.data)
			other.data.clear()
			other.data.extend(self.data)
	#

#



#
# This class implements a buffer for the text console. It consists of individual cells, one for every character.
#
class ConsoleBufferRGB2(object):

	################################################################################################################################
	## Constants
	################################################################################################################################

	################################################################################################################################
	## Constructor
	################################################################################################################################

	#
	# Constructor method.
	#
	def __init__(self, width:int = None, height:int = None, bgColor:int = None, bTransparent:bool = False):
		self.__width = Console.width() - 1 if width is None else width
		self.__height = Console.height() - 1 if height is None else height
		self.__widthM1 = self.__width - 1
		self.__heightM1 = self.__height - 1
		self.__numberOfCells = self.__width * self.__height
		self.__bTransparent = bTransparent
		self.__bgColor = bgColor

		self.__lineNumbers = tuple(range(0, self.__height))
		self.__columnNumbers = tuple(range(0, self.__width))

		c = None if bTransparent else " "
		self.__dataMatrix = [ [ _ConsoleBufferRGBCell2(bgColor, None, c) for x in self.__columnNumbers ] for y in self.__lineNumbers ]

		# ----------------------------------------------------------------

		self.__asserts()

		# ----------------------------------------------------------------
		# caching

		self.__colorMapper = RGBColorMapper()
		self.__oldData = None
		self.__oldX = None
		self.__oldY = None

		# ----------------------------------------------------------------

		self.defaultFramedBoxSettings = FramedBoxSettingsRGB()

		# ----------------------------------------------------------------

		# if bgColor is not None:
		# 	self.fill(bgColor, None, None)
	#

	################################################################################################################################
	## Public Properties
	################################################################################################################################

	#
	# Returns a tuple with two values: The width and height.
	#
	@property
	def size(self):
		return self.__width, self.__height
	#

	@property
	def width(self):
		return self.__width
	#

	@property
	def height(self):
		return self.__height
	#

	################################################################################################################################
	## Helper Methods
	################################################################################################################################

	def __asserts(self):
		assert len(self.__lineNumbers) == self.__height
		assert len(self.__columnNumbers) == self.__width
		assert len(self.__dataMatrix) == self.__height
		for i in range(0, self.__height):
			assert len(self.__dataMatrix[i]) == self.__width
	#

	def __dumpToSimpleTable(self) -> SimpleTable:
		t = SimpleTable()
		for row in self.__dataMatrix:
			tRow = t.addRow()
			for cell in row:
				d = [
					IntRGB.toCSS(cell.data[0]) if cell.data[0] is not None else "-",
					IntRGB.toCSS(cell.data[1]) if cell.data[1] is not None else "-",
					repr(cell.data[2]) ]
				tRow.addCell().value = ":".join(d)

		for i in range(0, t.numberOfColumns - 1):
			t.column(i).vlineAfterColumn = True
		for i in range(0, t.numberOfRows - 1):
			t.row(i).hlineAfterRow = True

		return t
	#

	################################################################################################################################
	## Public Methods
	################################################################################################################################

	################################################################################################################################
	## Public Static Methods
	################################################################################################################################

	#
	# Grow the buffer vertically by the specified number of rows.
	#
	def growBufferY(self, nRows:int):
		assert isinstance(nRows, int)
		assert nRows >= 0

		if nRows == 0:
			return

		c = None if self.__bTransparent else " "
		for i in range(0, nRows):
			_row = [ _ConsoleBufferRGBCell2(self.__bgColor, None, c) for x in self.__columnNumbers ]
			self.__dataMatrix.append(_row)

		self.__height += nRows

		self.__heightM1 = self.__height - 1
		self.__numberOfCells = self.__width * self.__height
		self.__lineNumbers = tuple(range(0, self.__height))

		self.__asserts()

		# clear caching

		self.__oldData = None
		self.__oldX = None
		self.__oldY = None
	#

	def extract(self, r:Rect):
		rf = r.intersect(Rect(0, 0, self.__width, self.__height))
		if (rf.width == 0) or (rf.height == 0):
			return None
		b = ConsoleBufferRGB2(rf.width, rf.height)
		self.bufferToBufferCopy(-r.x, -r.y, b)
		return b
	#

	#
	# During rendering to the console color values are mapped from int to str.
	# This cache can grow as it is not cleared automatically. If you use a large variety of colors
	# you might want to clear this cache from time to time.
	#
	def clearColorRenderingCache(self):
		self.__colorMapper.clear()
	#

	def __str__(self):
		return "<ConsoleBufferRGB2 size=%dx%d at %x>" % (self.__width, self.__height, id(self))
	#

	def __repr__(self):
		return "<ConsoleBufferRGB2 size=%dx%d at %x>" % (self.__width, self.__height, id(self))
	#

	#
	# Copy the content of this buffer overwriting all data in the cells of the destination buffer.
	#
	def bufferToBufferCopy(self, ofsX:int, ofsY:int, other):
		assert isinstance(other, ConsoleBufferRGB2)

		sourceRect = _Rect(0, 0, self.__width, self.__height)
		destRect = _Rect(0, 0, other.__width, other.__height)
		destRect0 = sourceRect.clone().move(ofsX, ofsY).intersect(destRect)
		sourceRect0 = destRect0.clone().move(-ofsX, -ofsY)

		for y in range(0, destRect0.height):
			currentRowSrc = self.__dataMatrix[sourceRect0.y + y]
			currentRowDest = other.__dataMatrix[destRect0.y + y]
			for x in range(0, destRect0.width):
				currentCellSrc = currentRowSrc[sourceRect0.x + x]
				currentCellDest = currentRowDest[destRect0.x + x]

				if currentCellDest.data != currentCellSrc.data:
					currentCellDest.data = list(currentCellSrc.data)
	#

	#
	# Copy the content of this buffer overwriting only data in the cells of the destination buffer the source buffer can provide.
	# This allows copying color information only if no content is present.
	#
	def bufferToBufferMerge(self, ofsX:int, ofsY:int, other):
		assert isinstance(other, ConsoleBufferRGB2)

		sourceRect = _Rect(0, 0, self.__width, self.__height)
		destRect = _Rect(0, 0, other.__width, other.__height)
		destRect0 = sourceRect.clone().move(ofsX, ofsY).intersect(destRect)
		sourceRect0 = destRect0.clone().move(-ofsX, -ofsY)

		for y in range(0, destRect0.height):
			currentRowSrc = self.__dataMatrix[sourceRect0.y + y]
			currentRowDest = other.__dataMatrix[destRect0.y + y]
			for x in range(0, destRect0.width):
				currentCellSrc = currentRowSrc[sourceRect0.x + x]
				currentCellDest = currentRowDest[destRect0.x + x]

				for i, v in enumerate(currentCellSrc.data):
					if v is not None:
						currentCellDest.data[i] = v
	#

	def bufferToConsole(self, ofsX:int = 0, ofsY:int = 0):
		if self.__bTransparent:
			raise Exception("Transparent buffers can not be copied to the screen. Rendering would fail.")

		# calculate change matrix and cache all data

		bChangedMatrix = []
		if (self.__oldData is not None) and (self.__oldX == ofsX) and (self.__oldY == ofsY):
			oldData = self.__oldData
			for iy, row in enumerate(self.__dataMatrix):
				bChangedRow = []
				for ix, cell in enumerate(row):
					bChanged = oldData[iy][ix] != cell.data
					bChangedRow.append(bChanged)
					if bChanged:
						oldData[iy][ix] = list(cell.data)
				bChangedMatrix.append(bChangedRow)
		else:
			oldData = []
			for iy, row in enumerate(self.__dataMatrix):
				bChangedRow = []
				oldRow = []
				for ix, cell in enumerate(row):
					bChangedRow.append(True)
					oldRow.append(list(cell.data))
				oldData.append(oldRow)
				bChangedMatrix.append(bChangedRow)
			self.__oldX = ofsX
			self.__oldY = ofsY
			self.__oldData = oldData

		# print the new data

		characters = []
		for y in self.__lineNumbers:
			currentRow = self.__dataMatrix[y]
			bCurrentRowChanged = bChangedMatrix[y]

			# scan for changed areas

			bLineIsModified = False
			fromX = 999999
			toX = -1
			for x in self.__columnNumbers:
				if bCurrentRowChanged[x]:
					fromX = x
					bLineIsModified = True
					break

			if bLineIsModified:
				for x in reversed(self.__columnNumbers):
					if bCurrentRowChanged[x]:
						toX = x
						break
			else:
				# no changes => continue
				continue

			# now perform paint

			characters.clear()
			characters.append("\033[" + str(ofsY + y + 1) + ";" + str(ofsX + fromX + 1) + "H")	# position cursor

			lastCol = None
			for x in self.__columnNumbers[fromX:toX + 1]:
				currentCell = currentRow[x]
				currentCol = (currentCell.data[0], currentCell.data[1])
				if lastCol != currentCol:	# has the color changed compared to last cell?
					# yes it has changed
					lastCol = currentCol
					characters.append(self.__colorMapper(currentCol[0], currentCol[1]))

				c = currentCell.data[2]
				characters.append(" " if c is None else c)

			characters.append(Console.RESET)
			print("".join(characters), end='', flush=False)

		print(Console.RESET_TOPLEFT, end="", flush=True)	# reset color output and position cursor
	#

	def dump(self):
		t = self.__dumpToSimpleTable()
		t.print()
	#

	def dumpToFile(self, filePath:str):
		assert isinstance(filePath, str)

		t = self.__dumpToSimpleTable()
		with open(filePath, "w") as fout:
			for line in t.printToLines():
				fout.write(line)
				fout.write("\n")
	#

	#
	# Fill the entire buffer with the specified color and character.
	#
	def fill(self, bgColor:int, fgColor:int, character:str = " "):
		if character is not None:
			assert isinstance(character, str)
			assert len(character) == 1
		if bgColor is not None:
			assert isinstance(bgColor, int)
		if fgColor is not None:
			assert isinstance(fgColor, int)

		for y in self.__lineNumbers:
			for x in self.__columnNumbers:
				cell = self.__dataMatrix[y][x]

				if bgColor is not None:
					cell.data[0] = bgColor

				if fgColor is not None:
					cell.data[1] = fgColor

				if character is not None:
					cell.data[2] = character
	#

	#
	# Fill a single cell with color and character.
	#
	def set(self, x:int, y:int, bgColor:int, fgColor:int, character:str):
		cell = self.__dataMatrix[y][x]

		if bgColor is not None:
			assert isinstance(bgColor, int)
			cell.data[0] = bgColor

		if fgColor is not None:
			assert isinstance(fgColor, int)
			cell.data[1] = fgColor

		if character is not None:
			assert isinstance(character, str)
			assert len(character) == 1
			cell.data[2] = character
	#

	#
	# Fill a single cell with color and character.
	#
	def setSafe(self, x:int, y:int, bgColor:int, fgColor:int, character:str):
		if (x < 0) or (y < 0) or (x >= self.__width) or (y >= self.__height):
			return

		cell = self.__dataMatrix[y][x]

		if bgColor is not None:
			assert isinstance(bgColor, int)
			cell.data[0] = bgColor

		if fgColor is not None:
			assert isinstance(fgColor, int)
			cell.data[1] = fgColor

		if character is not None:
			assert isinstance(character, str)
			assert len(character) == 1
			cell.data[2] = character
	#

	def printText(self, x:int, y:int, bgColor:int, fgColor:int, text:str):
		assert isinstance(x, int)
		assert isinstance(y, int)
		if bgColor is not None:
			assert isinstance(bgColor, int)
		if fgColor is not None:
			assert isinstance(fgColor, int)
		assert isinstance(text, str)
		if len(text) == 0:
			return

		for i, c in enumerate(text):
			cell = self.__dataMatrix[y][x + i]

			if bgColor is not None:
				cell.data[0] = bgColor

			if fgColor is not None:
				cell.data[1] = fgColor

			cell.data[2] = c
	#

	def hline(self, x1:int, y:int, x2:int, bgColor:int, fgColor:int, character:str):
		assert isinstance(x1, int)
		assert isinstance(y, int)
		assert isinstance(x2, int)
		if bgColor is not None:
			assert isinstance(bgColor, int)
		if fgColor is not None:
			assert isinstance(fgColor, int)
		assert isinstance(character, str)
		assert len(character) == 1

		row = self.__dataMatrix[y]
		for ix in range(x1, x2 + 1):
			cell = row[ix]

			if bgColor is not None:
				cell.data[0] = bgColor

			if fgColor is not None:
				cell.data[1] = fgColor

			if cell.data[2] != character:
				cell.data[2] = character
	#

	def vline(self, x:int, y1:int, y2:int, bgColor:int, fgColor:int, character:str):
		assert isinstance(x, int)
		assert isinstance(y1, int)
		assert isinstance(y2, int)
		if bgColor is not None:
			assert isinstance(bgColor, int)
		if fgColor is not None:
			assert isinstance(fgColor, int)
		assert isinstance(character, str)
		assert len(character) == 1

		for iy in range(y1, y2 + 1):
			cell = self.__dataMatrix[iy][x]

			if bgColor is not None:
				cell.data[0] = bgColor

			if fgColor is not None:
				cell.data[1] = fgColor

			if cell.data[2] != character:
				cell.data[2] = character
	#

	def drawFrame(self, rect:Rect, settings:FramedBoxSettingsRGB = None):
		assert isinstance(rect, Rect)
		x1 = rect.x1
		y1 = rect.y1
		x2 = rect.x2 - 1		# TODO
		y2 = rect.y2 - 1		# TODO

		if settings is None:
			settings = self.defaultFramedBoxSettings
		else:
			assert isinstance(settings, FramedBoxSettingsRGB)

		self.hline(x1, y1, x2, settings.frameColorBG, settings.frameColorFG, settings.frameCharTop)
		self.hline(x1, y2, x2, settings.frameColorBG, settings.frameColorFG, settings.frameCharBottom)
		self.vline(x1, y1, y2, settings.frameColorBG, settings.frameColorFG, settings.frameCharLeft)
		self.vline(x2, y1, y2, settings.frameColorBG, settings.frameColorFG, settings.frameCharRight)
		self.set(x1, y1, settings.frameColorBG, settings.frameColorFG, settings.frameCharTopLeft)
		self.set(x2, y1, settings.frameColorBG, settings.frameColorFG, settings.frameCharTopRight)
		self.set(x1, y2, settings.frameColorBG, settings.frameColorFG, settings.frameCharBottomLeft)
		self.set(x2, y2, settings.frameColorBG, settings.frameColorFG, settings.frameCharBottomRight)
	#

	def fillRectangle(self, rect:Rect, bgColor:int, fgColor:int, character:str):
		assert isinstance(rect, Rect)
		x1 = rect.x1
		y1 = rect.y1
		x2 = rect.x2		# TODO
		y2 = rect.y2		# TODO

		if bgColor is not None:
			assert isinstance(bgColor, int)

		if fgColor is not None:
			assert isinstance(fgColor, int)

		assert isinstance(character, str)
		assert len(character) == 1

		for iy in range(y1, y2):
			row = self.__dataMatrix[iy]
			for ix in range(x1, x2):
				cell = row[ix]

				if bgColor is not None:
					cell.data[0] = bgColor

				if fgColor is not None:
					cell.data[1] = fgColor

				if cell.data[2] != character:
					cell.data[2] = character
	#

#









from ..Console import Console




class _RendererImpl(object):

	MAX_BRIGHTNESS = 20

	def __init__(self):
		self.bgTables = []
		self.fgTables = []
		for i in range(0, 101, 5):
			f = i / 100
			self.bgTables.append(_RendererImpl.__createRGBTableBG(f))
			self.fgTables.append(_RendererImpl.__createRGBTableFG(f))
	#

	@staticmethod
	def __createRGBTableBG(brightness:float):
		assert 0 <= brightness <= 1

		f = brightness * 16
		ret = [ 0 ] * (16 * 16 * 16)
		for r in range(0, 16):
			for g in range(0, 16):
				for b in range(0, 16):
					index = r * 256 + g * 16 + b
					ret[index] = "\033[48;2;" + str(int(r * f)) + ";" + str(int(g * f)) + ";" + str(int(b * f)) + "m"
		return ret
	#

	@staticmethod
	def __createRGBTableFG(brightness:float):
		assert 0 <= brightness <= 1

		f = brightness * 16
		ret = [ 0 ] * (16 * 16 * 16)
		for r in range(0, 16):
			for g in range(0, 16):
				for b in range(0, 16):
					index = r * 256 + g * 16 + b
					ret[index] = "\033[38;2;" + str(int(r * f)) + ";" + str(int(g * f)) + ";" + str(int(b * f)) + "m"
		return ret
	#

	"""
	def toConsoleStr(self, c:str, bgIndex:int, fgIndex:int, brightness:int):
		ret = []
		if bgIndex is not None:
			ret.append(self.bgTables[brightness][bgIndex])
		if fgIndex is not None:
			ret.append(self.bgTables[brightness][bgIndex])
		if c is not None:
			ret.append(c)
		return self.bgTables[brightness][bgIndex] + self.fgTables[brightness][fgIndex] + c
	#
	"""

#

ViewPortRenderer = _RendererImpl()













__slots__ = (
	"parseCSS_toARGB",
	"parseCSS_to3Ints",
)


import typing
import re






_HEXVALS = "0123456789abcdef"
_P1 = re.compile(r"^#([0-9a-f][0-9a-f])([0-9a-f][0-9a-f])([0-9a-f][0-9a-f])$")
_P2 = re.compile(r"^#([0-9a-f])([0-9a-f])([0-9a-f])$")
_P3 = re.compile(r"^rgb\(\s*([0-9]+)\s*,\s*([0-9]+)\s*,\s*([0-9]+)\s*\)$")







def parseCSS_toARGB(cssStr:str) -> int:
	assert isinstance(cssStr, str)

	cssStr = cssStr.lower()
	m = _P1.match(cssStr)
	if m:
		sr = m.group(1)
		sg = m.group(2)
		sb = m.group(3)
		r = (_HEXVALS.index(sr[0]) << 4) + _HEXVALS.index(sr[1])
		g = (_HEXVALS.index(sg[0]) << 4) + _HEXVALS.index(sg[1])
		b = (_HEXVALS.index(sb[0]) << 4) + _HEXVALS.index(sb[1])
		return 0xff000000 | ((((r << 8) + g) << 8) + b)
	m = _P2.match(cssStr)
	if m:
		sr = m.group(1)
		sg = m.group(2)
		sb = m.group(3)
		r = _HEXVALS.index(sr)
		g = _HEXVALS.index(sg)
		b = _HEXVALS.index(sb)
		return 0xff000000 | ((((((((((r << 4) + r) << 4) + g) << 4) + g) << 4) + b) << 4) + b)
	m = _P3.match(cssStr)
	if m:
		sr = m.group(1)
		sg = m.group(2)
		sb = m.group(3)
		r = int(sr)
		if r > 255:
			raise Exception("Invalid red component in: " + repr(cssStr))
		g = int(sg)
		if g > 255:
			raise Exception("Invalid green component in: " + repr(cssStr))
		b = int(sb)
		if b > 255:
			raise Exception("Invalid blue component in: " + repr(cssStr))
		return 0xff000000 | ((((r << 8) + g) << 8) + b)
	raise Exception("Not a CSS color string: " + repr(cssStr))
#



def parseCSS_to3Ints(cssStr:str) -> typing.Tuple[int,int,int]:
	assert isinstance(cssStr, str)

	cssStr = cssStr.lower()
	m = _P1.match(cssStr)
	if m:
		sr = m.group(1)
		sg = m.group(2)
		sb = m.group(3)
		r = (_HEXVALS.index(sr[0]) << 4) + _HEXVALS.index(sr[1])
		g = (_HEXVALS.index(sg[0]) << 4) + _HEXVALS.index(sg[1])
		b = (_HEXVALS.index(sb[0]) << 4) + _HEXVALS.index(sb[1])
		return (r, g, b)
	m = _P2.match(cssStr)
	if m:
		sr = m.group(1)
		sg = m.group(2)
		sb = m.group(3)
		r = _HEXVALS.index(sr)
		g = _HEXVALS.index(sg)
		b = _HEXVALS.index(sb)
		return ((r << 4) + r, (g << 4) + g, (b << 4) + b)
	m = _P3.match(cssStr)
	if m:
		sr = m.group(1)
		sg = m.group(2)
		sb = m.group(3)
		r = int(sr)
		if r > 255:
			raise Exception("Invalid red component in: " + repr(cssStr))
		g = int(sg)
		if g > 255:
			raise Exception("Invalid green component in: " + repr(cssStr))
		b = int(sb)
		if b > 255:
			raise Exception("Invalid blue component in: " + repr(cssStr))
		return (r, g, b)
	raise Exception("Not a CSS color string: " + repr(cssStr))
#









from .Console import *
from .ConsoleBuffer import *
from .Rect import *






class CharacterBuffer(object):

	class _Cell(object):

		def __init__(self, bgColor:str):
			self.col = bgColor
			self.c = " "
		#

		def clear(self, bgColor:str):
			self.col = bgColor
			self.c = " "
		#

	#

	def __init__(self, width:int, height:int, bgColor:str = Console.RESET):
		self.__bgColor = bgColor
		self.__width = width
		self.__height = height
		self.__widthM1 = self.__width - 1
		self.__heightM1 = self.__height - 1
		self.__numberOfCells = self.__width * self.__height

		self.__lineNumbers = tuple(range(0, self.__height))
		self.__columnNumbers = tuple(range(0, self.__width))

		self._currentBuffer = [ [CharacterBuffer._Cell(bgColor) for x in self.__columnNumbers] for y in self.__lineNumbers]
	#

	@property
	def bounds(self):
		return Rect(0, 0, self.__width, self.__height)
	#

	@property
	def size(self):
		return self.__width, self.__height
	#

	@property
	def width(self):
		return self.__width
	#

	@property
	def height(self):
		return self.__height
	#

	def clear(self):
		bgColor = self.__bgColor
		for y in self.__lineNumbers:
			for x in self.__columnNumbers:
				self._currentBuffer[y][x].clear(bgColor)
	#

	def set(self, x:int, y:int, color:str, character:str):
		assert len(character) == 1

		cell = self._currentBuffer[y][x]
		if color != None:
			cell.c = character
		if color != None:
			cell.col = color
	#

	def fill(self, x:int, y:int, w:int, h:int, color:str, character:str):
		assert len(character) == 1

		for iy in range(y, y + h):
			row = self._currentBuffer[iy]
			for ix in range(x, x + w):
				cell = row[ix]
				cell.c = character
				if color != None:
					cell.col = color
	#

	def printText(self, x:int, y:int, color:str, text:str):
		assert len(text) > 0

		i = 0
		for c in text:
			cell = self._currentBuffer[y][x + i]
			cell.c = c
			if color != None:
				cell.col = color
			i += 1
	#

	def copyToCharacterBuffer1(self, srcX:int, srcY:int, width:int, height:int, destX:int, destY:int, otherBuffer):
		assert isinstance(otherBuffer, CharacterBuffer)

		rSrc = Rect(srcX, srcY, width, height)
		assert self.bounds.isOtherRectInRect(rSrc)
		rDest = Rect(destX, destY, width, height)
		assert otherBuffer.bounds.isOtherRectInRect(rDest)

		deltaX = destX - srcX
		deltaY = destY - srcY

		# ----------------

		for srcIY in range(srcY, srcY + height):
			srcRow = self._currentBuffer[srcIY]
			destRow = otherBuffer._currentBuffer[srcIY + deltaY]
			for srcIX in range(srcX, srcX + width):
				srcCell = srcRow[srcIX]
				destCell = destRow[srcIX + deltaX]
				destCell.c = srcCell.c
				destCell.col = srcCell.col
	#

	def copyToCharacterBuffer0(self, destX:int, destY:int, otherBuffer):
		assert isinstance(otherBuffer, CharacterBuffer)

		width = self.__width
		height = self.__height

		rDest = Rect(destX, destY, width, height)
		assert otherBuffer.bounds.isOtherRectInRect(rDest)

		deltaX = destX
		deltaY = destY
		srcX = 0
		srcY = 0

		# ----------------

		for srcIY in range(srcY, srcY + height):
			srcRow = self._currentBuffer[srcIY]
			destRow = otherBuffer._currentBuffer[srcIY + deltaY]
			for srcIX in range(srcX, srcX + width):
				srcCell = srcRow[srcIX]
				destCell = destRow[srcIX + deltaX]
				destCell.c = srcCell.c
				destCell.col = srcCell.col
	#

	def copyToConsoleBuffer1(self, srcX:int, srcY:int, width:int, height:int, destX:int, destY:int, otherBuffer:ConsoleBuffer):
		rSrc = Rect(srcX, srcY, width, height)
		assert self.bounds.isOtherRectInRect(rSrc)
		rDest = Rect(destX, destY, width, height)
		assert otherBuffer.bounds.isOtherRectInRect(rDest)

		deltaX = destX - srcX
		deltaY = destY - srcY

		# ----------------

		for srcIY in range(srcY, srcY + height):
			srcRow = self._currentBuffer[srcIY]
			destRow = otherBuffer._currentBuffer[srcIY + deltaY]
			for srcIX in range(srcX, srcX + width):
				srcCell = srcRow[srcIX]
				destCell = destRow[srcIX + deltaX]
				destCell.c = srcCell.c
				destCell.col = srcCell.col
				destCell.bChanged = True
	#

	def copyToConsoleBuffer0(self, destX:int, destY:int, otherBuffer:ConsoleBuffer):
		width = self.__width
		height = self.__height

		rDest = Rect(destX, destY, width, height)
		assert otherBuffer.bounds.isOtherRectInRect(rDest)

		deltaX = destX
		deltaY = destY
		srcX = 0
		srcY = 0

		# ----------------

		for srcIY in range(srcY, srcY + height):
			srcRow = self._currentBuffer[srcIY]
			destRow = otherBuffer._currentBuffer[srcIY + deltaY]
			for srcIX in range(srcX, srcX + width):
				srcCell = srcRow[srcIX]
				destCell = destRow[srcIX + deltaX]
				destCell.c = srcCell.c
				destCell.col = srcCell.col
				destCell.bChanged = True
	#

#









class RGBColorMapper(object):

	def __init__(self):
		# for rendering to the console: these maps hold int->str pairs assigning the string representations of colors to their integer representations
		self.__colMapBG = {}
		self.__colMapFG = {}
	#

	#
	# During rendering to the console color values are mapped from int to str.
	# This cache can grow as it is not cleared automatically. If you use a large variety of colors
	# you might want to clear this cache from time to time.
	#
	def clear(self):
		self.__colMapBG.clear()
		self.__colMapFG.clear()
	#

	#
	# Convert the rgb color values to a string. The result is cached.
	#
	def get(self, bgRGB:int, fgRGB:int):
		s = ""
		if bgRGB is not None:
			colStr = self.__colMapBG.get(bgRGB)
			if colStr is None:
				colStr = "\033[48;2;" + str((bgRGB >> 16) & 0xff) + ";" + str((bgRGB >> 8) & 0xff) + ";" + str(bgRGB & 0xff) + "m"
				self.__colMapBG[bgRGB] = colStr
			s += colStr
		if fgRGB is not None:
			colStr = self.__colMapFG.get(fgRGB)
			if colStr is None:
				colStr = "\033[38;2;" + str((fgRGB >> 16) & 0xff) + ";" + str((fgRGB >> 8) & 0xff) + ";" + str(fgRGB & 0xff) + "m"
				self.__colMapFG[fgRGB] = colStr
			s += colStr
		return s
	#

	#
	# Convert the rgb color values to a string. The result is cached.
	#
	def __call__(self, bgRGB:int, fgRGB:int):
		s = ""
		if bgRGB is not None:
			colStr = self.__colMapBG.get(bgRGB)
			if colStr is None:
				colStr = "\033[48;2;" + str((bgRGB >> 16) & 0xff) + ";" + str((bgRGB >> 8) & 0xff) + ";" + str(bgRGB & 0xff) + "m"
				self.__colMapBG[bgRGB] = colStr
			s += colStr
		if fgRGB is not None:
			colStr = self.__colMapFG.get(fgRGB)
			if colStr is None:
				colStr = "\033[38;2;" + str((fgRGB >> 16) & 0xff) + ";" + str((fgRGB >> 8) & 0xff) + ";" + str(fgRGB & 0xff) + "m"
				self.__colMapFG[fgRGB] = colStr
			s += colStr
		return s
	#

#


















from .Rectangle import Rectangle
from .ViewPortRGB import ViewPortRGB
from .ViewPortCell import ViewPortCell
from .ViewPortBuffer import ViewPortBuffer
from .ViewPortBuffer2 import ViewPortBuffer2
from .ViewPort import ViewPort

from .ViewPortRenderer import ViewPortRenderer as __vpr
MAX_BRIGHTNESS = __vpr.MAX_BRIGHTNESS







from .Console import Console as _Console
_ConsoleFG = _Console.ForeGround




def bold(text:str) -> str:
	return "\x1b[1m" + text + "\x1b[22m"
#

def underline(text:str) -> str:
	return "\x1b[4m" + text + "\x1b[24m"
#

def italic(text:str) -> str:
	return "\x1b[3m" + text + "\x1b[23m"
#



def stdBlack(text:str) -> str:
	return _ConsoleFG.STD_BLACK + text + "\x1b[39m"
#

def stdBlue(text:str) -> str:
	return _ConsoleFG.STD_BLUE + text + "\x1b[39m"
#

def stdGreen(text:str) -> str:
	return _ConsoleFG.STD_GREEN + text + "\x1b[39m"
#

def stdCyan(text:str) -> str:
	return _ConsoleFG.STD_CYAN + text + "\x1b[39m"
#

def stdRed(text:str) -> str:
	return _ConsoleFG.STD_RED + text + "\x1b[39m"
#

def stdPurple(text:str) -> str:
	return _ConsoleFG.STD_PURPLE + text + "\x1b[39m"
#

def stdDarkYellow(text:str) -> str:
	return _ConsoleFG.STD_DARKYELLOW + text + "\x1b[39m"
#

def stdLightGray(text:str) -> str:
	return _ConsoleFG.STD_LIGHTGRAY + text + "\x1b[39m"
#

def stdDarkGray(text:str) -> str:
	return _ConsoleFG.STD_DARKGRAY + text + "\x1b[39m"
#

def stdLightBlue(text:str) -> str:
	return _ConsoleFG.STD_LIGHTBLUE + text + "\x1b[39m"
#

def stdLightGreen(text:str) -> str:
	return _ConsoleFG.STD_LIGHTGREEN + text + "\x1b[39m"
#

def stdLightCyan(text:str) -> str:
	return _ConsoleFG.STD_LIGHTCYAN + text + "\x1b[39m"
#

def stdLightRed(text:str) -> str:
	return _ConsoleFG.STD_LIGHTRED + text + "\x1b[39m"
#

def stdLightPurple(text:str) -> str:
	return _ConsoleFG.STD_LIGHTPURPLE + text + "\x1b[39m"
#

def stdYellow(text:str) -> str:
	return _ConsoleFG.STD_YELLOW + text + "\x1b[39m"
#

def stdWhite(text:str) -> str:
	return _ConsoleFG.STD_WHITE + text + "\x1b[39m"
#



def black(text:str) -> str:
	return _ConsoleFG.BLACK + text + "\x1b[39m"
#

def darkGray(text:str) -> str:
	return _ConsoleFG.DARK_GRAY + text + "\x1b[39m"
#

def gray(text:str) -> str:
	return _ConsoleFG.GRAY + text + "\x1b[39m"
#

def lightGray(text:str) -> str:
	return _ConsoleFG.LIGHT_GRAY + text + "\x1b[39m"
#

def white(text:str) -> str:
	return _ConsoleFG.WHITE + text + "\x1b[39m"
#

def red(text:str) -> str:
	return _ConsoleFG.RED + text + "\x1b[39m"
#

def orange(text:str) -> str:
	return _ConsoleFG.ORANGE + text + "\x1b[39m"
#

def yellow(text:str) -> str:
	return _ConsoleFG.YELLOW + text + "\x1b[39m"
#

def yellowGreen(text:str) -> str:
	return _ConsoleFG.YELLOWGREEN + text + "\x1b[39m"
#

def green(text:str) -> str:
	return _ConsoleFG.GREEN + text + "\x1b[39m"
#

def greenCyan(text:str) -> str:
	return _ConsoleFG.GREENCYAN + text + "\x1b[39m"
#

def cyan(text:str) -> str:
	return _ConsoleFG.CYAN + text + "\x1b[39m"
#

def cyanBlue(text:str) -> str:
	return _ConsoleFG.CYANBLUE + text + "\x1b[39m"
#

def blue(text:str) -> str:
	return _ConsoleFG.BLUE + text + "\x1b[39m"
#

def blueViolet(text:str) -> str:
	return _ConsoleFG.BLUEVIOLET + text + "\x1b[39m"
#

def violet(text:str) -> str:
	return _ConsoleFG.VIOLET + text + "\x1b[39m"
#

def violetRed(text:str) -> str:
	return _ConsoleFG.VIOLETRED + text + "\x1b[39m"
#







def darkRed(text:str) -> str:
	return _ConsoleFG.DARK_RED + text + "\x1b[39m"
#

def darkOrange(text:str) -> str:
	return _ConsoleFG.DARK_ORANGE + text + "\x1b[39m"
#

def darkYellow(text:str) -> str:
	return _ConsoleFG.DARK_YELLOW + text + "\x1b[39m"
#

def darkYellowGreen(text:str) -> str:
	return _ConsoleFG.DARK_YELLOWGREEN + text + "\x1b[39m"
#

def darkGreen(text:str) -> str:
	return _ConsoleFG.DARK_GREEN + text + "\x1b[39m"
#

def darkGreenCyan(text:str) -> str:
	return _ConsoleFG.DARK_GREENCYAN + text + "\x1b[39m"
#

def darkCyan(text:str) -> str:
	return _ConsoleFG.DARK_CYAN + text + "\x1b[39m"
#

def darkCyanBlue(text:str) -> str:
	return _ConsoleFG.DARK_CYANBLUE + text + "\x1b[39m"
#

def darkBlue(text:str) -> str:
	return _ConsoleFG.DARK_BLUE + text + "\x1b[39m"
#

def darkBlueViolet(text:str) -> str:
	return _ConsoleFG.DARK_BLUEVIOLET + text + "\x1b[39m"
#

def darkViolet(text:str) -> str:
	return _ConsoleFG.DARK_VIOLET + text + "\x1b[39m"
#

def darkVioletRed(text:str) -> str:
	return _ConsoleFG.DARK_VIOLETRED + text + "\x1b[39m"
#







def lightRed(text:str) -> str:
	return _ConsoleFG.LIGHT_RED + text + "\x1b[39m"
#

def lightOrange(text:str) -> str:
	return _ConsoleFG.LIGHT_ORANGE + text + "\x1b[39m"
#

def lightYellow(text:str) -> str:
	return _ConsoleFG.LIGHT_YELLOW + text + "\x1b[39m"
#

def lightYellowGreen(text:str) -> str:
	return _ConsoleFG.LIGHT_YELLOWGREEN + text + "\x1b[39m"
#

def lightGreen(text:str) -> str:
	return _ConsoleFG.LIGHT_GREEN + text + "\x1b[39m"
#

def lightGreenCyan(text:str) -> str:
	return _ConsoleFG.LIGHT_GREENCYAN + text + "\x1b[39m"
#

def lightCyan(text:str) -> str:
	return _ConsoleFG.LIGHT_CYAN + text + "\x1b[39m"
#

def lightCyanBlue(text:str) -> str:
	return _ConsoleFG.LIGHT_CYANBLUE + text + "\x1b[39m"
#

def lightBlue(text:str) -> str:
	return _ConsoleFG.LIGHT_BLUE + text + "\x1b[39m"
#

def lightBlueViolet(text:str) -> str:
	return _ConsoleFG.LIGHT_BLUEVIOLET + text + "\x1b[39m"
#

def lightViolet(text:str) -> str:
	return _ConsoleFG.LIGHT_VIOLET + text + "\x1b[39m"
#

def lightVioletRed(text:str) -> str:
	return _ConsoleFG.LIGHT_VIOLETRED + text + "\x1b[39m"
#








from typing import Union

from ..Console import Console
from .ViewPortCell import ViewPortCell
from .ViewPortRenderer import ViewPortRenderer
from .ViewPortRGB import ViewPortRGB
from .Rectangle import Rectangle






class ViewPortBuffer2(object):

	LINE_STYLE_SIMPLE = [ "-", "|", "/", "\\", "\\", "/" ]

	def __init__(self, viewPort, ofsX:int, ofsY:int, width:int, height:int):
		self.__viewPort = viewPort
		self.__width = width
		self.__height = height
		self.__ofsX = ofsX
		self.__ofsY = ofsY
		self.rows = []
		for y in range(0, height):
			row = [ None ] * self.__width
			self.rows.append(row)
			for x in range(0, width):
				row[x] = ViewPortCell()
		self.__brightness = ViewPortRenderer.MAX_BRIGHTNESS
		self._bModified = False
		self.bErrorIfOutOfBounds = False
	#

	def _setBrightness(self, brightness:int):
		for row in self.rows:
			for cell in row:
				cell._brightness = brightness
		self.__brightness = brightness
	#

	@property
	def viewPort(self):
		return self.__viewPort
	#

	@property
	def height(self) -> int:
		return self.__height
	#

	@property
	def width(self) -> int:
		return self.__width
	#

	@property
	def ofsX(self) -> int:
		return self.__ofsX
	#

	@property
	def ofsY(self) -> int:
		return self.__ofsY
	#

	@property
	def bounds(self) -> Rectangle:
		return Rectangle(self.__ofsX, self.__ofsY, self.__width, self.__height)
	#

	@property
	def sizeRect(self) -> Rectangle:
		return Rectangle(0, 0, self.__width, self.__height)
	#

	def drawText(self, *args, text:str = None, colBG:ViewPortRGB = None, colFG:ViewPortRGB = None):
		if len(args) == 1:
			assert isinstance(args[0], (tuple, list))
			x = args[0][0]
			y = args[0][1]
		elif len(args) == 2:
			x = args[0]
			y = args[1]
		else:
			raise Exception("args ???")

		assert isinstance(text, str)

		assert isinstance(x, int)
		assert isinstance(y, int)

		if self.bErrorIfOutOfBounds:
			row = self.rows[y]
			for i, c in enumerate(text[0:self.__width]):
				cell = row[x + i]
				if c is not None:
					cell.c = c
				if colBG is not None:
					cell.colBG = int(colBG)
				if colFG is not None:
					cell.colFG = int(colFG)

		else:
			if 0 <= y < self.__height:
				row = self.rows[y]
				for i, c in enumerate(text[0:self.__width]):
					#self.rows[y][x + i].set(c, colBG, colFG)

					ix = x + i
					if 0 <= ix < self.__width:
						cell = row[ix]
						if c is not None:
							cell.c = c
						if colBG is not None:
							cell.colBG = int(colBG)
						if colFG is not None:
							cell.colFG = int(colFG)
	#

	def set(self, x:int, y:int, c:str = None, colBG:ViewPortRGB = None, colFG:ViewPortRGB = None):
		#if len(args) == 1:
		#	assert isinstance(args[0], (tuple, list))
		#	x = args[0][0]
		#	y = args[0][1]
		#elif len(args) == 2:
		#	x = args[0]
		#	y = args[1]
		#else:
		#	raise Exception("args ???")

		#assert isinstance(c, str)

		#assert isinstance(x, int)
		#assert isinstance(y, int)

		# self.rows[y][x].set(c, colBG, colFG)

		if self.bErrorIfOutOfBounds:
			cell = self.rows[y][x]
			if c is not None:
				cell.c = c
			if colBG is not None:
				cell.colBG = int(colBG)
			if colFG is not None:
				cell.colFG = int(colFG)

		else:
			if (0 <= y < self.__height) and (0 <= x < self.__width):
				cell = self.rows[y][x]
				if c is not None:
					cell.c = c
				if colBG is not None:
					cell.colBG = int(colBG)
				if colFG is not None:
					cell.colFG = int(colFG)
	#

	def drawRect(self, *args, colBG:ViewPortRGB = None, colFG:ViewPortRGB = None, lineStyle = None):
		if len(args) == 1:
			r = args[0]
			if isinstance(r, Rectangle):
				x1 = r.x1
				y1 = r.y1
				w = r.width
				h = r.height
			elif isinstance(r, (tuple, list)):
				assert len(r) == 4
				x1 = r[0]
				y1 = r[1]
				w = r[2]
				h = r[3]
			else:
				raise Exception("args ???")
		elif len(args) == 4:
			x1 = args[0]
			y1 = args[1]
			w = args[2]
			h = args[3]
		else:
			raise Exception("args ???")

		assert isinstance(x1, int)
		assert isinstance(y1, int)
		assert isinstance(w, int)
		assert isinstance(h, int)

		if lineStyle is None:
			lineStyle = ViewPortBuffer2.LINE_STYLE_SIMPLE
		else:
			assert isinstance(lineStyle, (tuple, list))
			assert len(lineStyle) == 6
		if colFG:
			assert isinstance(colFG, (int, ViewPortRGB))
		if colBG:
			assert isinstance(colBG, (int, ViewPortRGB))

		hh, vv, tl, tr, bl, br = lineStyle
		x2 = x1 + w - 1
		y2 = y1 + h - 1

		if self.bErrorIfOutOfBounds:
			self.rows[y1][x1].set(tl, colBG, colFG)
			self.rows[y1][x2].set(tr, colBG, colFG)
			self.rows[y2][x1].set(bl, colBG, colFG)
			self.rows[y2][x2].set(br, colBG, colFG)
			for x in range(x1 + 1, x2):
				self.rows[y1][x].set(hh, colBG, colFG)
				self.rows[y2][x].set(hh, colBG, colFG)
			for y in range(y1 + 1, y2):
				self.rows[y][x1].set(vv, colBG, colFG)
				self.rows[y][x2].set(vv, colBG, colFG)

		else:
			if 0 <= y1 < self.__height:
				if 0 <= x1 < self.__width:
					self.rows[y1][x1].set(tl, colBG, colFG)
				if 0 <= x2 < self.__width:
					self.rows[y1][x2].set(tr, colBG, colFG)
			if 0 <= y2 < self.__height:
				if 0 <= x1 < self.__width:
					self.rows[y2][x1].set(bl, colBG, colFG)
				if 0 <= x2 < self.__width:
					self.rows[y2][x2].set(br, colBG, colFG)
			for x in range(x1 + 1, x2):
				if 0 <= x < self.__width:
					if 0 <= y1 < self.__height:
						self.rows[y1][x].set(hh, colBG, colFG)
					if 0 <= y2 < self.__height:
						self.rows[y2][x].set(hh, colBG, colFG)
			for y in range(y1 + 1, y2):
				if 0 <= y < self.__height:
					if 0 <= x1 < self.__width:
						self.rows[y][x1].set(vv, colBG, colFG)
					if 0 <= x2 < self.__width:
						self.rows[y][x2].set(vv, colBG, colFG)
	#

	def fill(self, c:str = None, colBG:ViewPortRGB = None, colFG:ViewPortRGB = None):
		if c is not None:
			assert isinstance(c, str)
			assert len(c) == 1
		if colFG:
			assert isinstance(colFG, (int, ViewPortRGB))
		if colBG:
			assert isinstance(colBG, (int, ViewPortRGB))

		for y in range(0, self.__height):
			for x in range(0, self.__width):
				cell = self.rows[y][x]
				if c is not None:
					cell.c = c
				if colBG:
					cell.colBG = int(colBG)
				if colFG:
					cell.colFG = int(colFG)

		self._bModified = True
	#

	def fillRect(self, x:int, y:int, w:int, h:int, c:str = None, colBG:ViewPortRGB = None, colFG:ViewPortRGB = None):
		assert isinstance(x, int)
		assert isinstance(y, int)
		assert isinstance(w, int)
		assert isinstance(h, int)

		if c is not None:
			assert isinstance(c, str)
			assert len(c) == 1
		if colFG:
			assert isinstance(colFG, (int, ViewPortRGB))
		if colBG:
			assert isinstance(colBG, (int, ViewPortRGB))

		if self.bErrorIfOutOfBounds:
			for iy in range(y, y+h):
				for ix in range(0, x+w):
					cell = self.rows[iy][ix]
					if c is not None:
						cell.c = c
					if colBG:
						cell.colBG = int(colBG)
					if colFG:
						cell.colFG = int(colFG)

		else:
			for iy in range(y, y+h):
				if 0 <= iy < self.__height:
					row = self.rows[iy]
					for ix in range(0, x+w):
						if 0 <= ix < self.__width:
							cell = row[ix]
							if c is not None:
								cell.c = c
							if colBG:
								cell.colBG = int(colBG)
							if colFG:
								cell.colFG = int(colFG)

		self._bModified = True
	#

#




















__author__ = "Jürgen Knauth"
__version__ = "0.2025.2.23.1"
__license__ = "Apache2"



from .Rect import Rect
from .Console import Console
from .ConsoleGraphics import ConsoleGraphics	# console buffer: cells with color ASCII string, text character and modification-flag; the buffer itself is without offset;
from .ConsoleBuffer import ConsoleBuffer		# console buffer: cells with color ASCII string, text character and modification-flag; the buffer itself has an offset value stored internall;
from .ConsoleBufferWO import ConsoleBufferWO	# simple version of ConsoleBuffer without modification flag
from .CharacterBuffer import CharacterBuffer	# int-based RGB values
from .IntRGB import IntRGB
from .ConsoleBufferRGB import ConsoleBufferRGB, FramedBoxSettingsRGB
from .SimpleTable import SimpleTable, SimpleTableCell, SimpleTableColumn, SimpleTableRow, SimpleTableConstants
from .EventCollector import EventCollector
from .FramedBoxSettingsRGB import FramedBoxSettingsRGB
from .ConsoleBufferRGB2 import ConsoleBufferRGB2

from . import wrappers










import typing

from .impl._parseCSS import parseCSS_toARGB







class IntRGB(object):

	#
	# Parse CSS codes as used by web technologies
	#
	@staticmethod
	def parseCSS(css:str):
		return parseCSS_toARGB(css)
	#

	@staticmethod
	def toCSS(rgb:int) -> str:
		assert isinstance(rgb, int)

		r = (rgb // 65536) % 256
		g = (rgb // 256) % 256
		b = rgb % 256
		return "#%0x%0x%0x" % (r, g, b)
	#

	@staticmethod
	def fade(rgbA:int, rgbB:int, f:typing.Union[float,int] = 0.5) -> int:
		assert isinstance(rgbA, int)
		assert isinstance(rgbB, int)
		assert isinstance(f, (float,int))

		rA = (rgbA // 65536) % 256
		gA = (rgbA // 256) % 256
		bA = rgbA % 256

		rB = (rgbB // 65536) % 256
		gB = (rgbB // 256) % 256
		bB = rgbB % 256

		if f < 0:
			f = 0
		if f > 1:
			f = 1
		fm1 = 1 - f

		r = round(rA*f + rB*fm1)
		g = round(gA*f + gB*fm1)
		b = round(bA*f + bB*fm1)

		return r * 65536 + g * 256 + b
	#

	@staticmethod
	def rgb256(r:int, g:int, b:int):
		assert isinstance(r, int)
		assert isinstance(g, int)
		assert isinstance(b, int)

		if not isinstance(r, int) or (r < 0) or (r > 255):
			raise Exception("Red value must be a valid integer value! (Value specified: " + str(r) + ")")
		if not isinstance(g, int) or (g < 0) or (g > 255):
			raise Exception("Red value must be a valid integer value! (Value specified: " + str(g) + ")")
		if not isinstance(b, int) or (b < 0) or (b > 255):
			raise Exception("Red value must be a valid integer value! (Value specified: " + str(b) + ")")
		return 0xff000000 | ((((r << 8) + g) << 8) + b)
	#

	@staticmethod
	def rgb1(r:float, g:float, b:float):
		if not isinstance(g, (int, float)) or (r < 0) or (r > 1):
			raise Exception("Red value must be a valid float value in the range [0..1]! (Value specified: " + str(r) + ")")
		if not isinstance(g, (int, float)) or (g < 0) or (g > 1):
			raise Exception("Green value must be a valid float value in the range [0..1]! (Value specified: " + str(g) + ")")
		if not isinstance(g, (int, float)) or (b < 0) or (b > 1):
			raise Exception("Blue value must be a valid float value in the range [0..1]! (Value specified: " + str(b) + ")")
		return 0xff000000 | ((((int(r * 255) << 8) + int(g * 255)) << 8) + int(b * 255))
	#

	@staticmethod
	def hsl1(h:float, s:float, l:float):
		if not isinstance(h, (int, float)) or (h < 0) or (h > 1):
			raise Exception("Hue value must be a valid float value in the range [0..1]! (h is " + str(h) + ")")
		if not isinstance(s, (int, float)) or (s < 0) or (s > 1):
			raise Exception("Saturation value must be a valid float value in the range [0..1]! (s is " + str(s) + ")")
		if not isinstance(l, (int, float)) or (l < 0) or (l > 1):
			raise Exception("Luminance value must be a valid float value in the range [0..1]! (l is " + str(l) + ")")

		if s == 0:
			s = 0.0001
		q = l * (1 + s) if l < 0.5 else l + s - l * s
		p = 2 * l - q
		r = IntRGB._hue2rgb(p, q, h + 1/3)
		g = IntRGB._hue2rgb(p, q, h)
		b = IntRGB._hue2rgb(p, q, h - 1/3)
		return 0xff000000 | (((int(r * 255) << 8) + int(g * 255)) << 8) + int(b * 255)
	#

	@staticmethod
	def _hue2rgb(p, q, t):
		if t < 0:
			t += 1
		if t > 1:
			t -= 1
		if t < 1/6:
			return p + (q - p) * 6 * t
		if t < 1/2:
			return q
		if t < 2/3:
			return p + (q - p) * (2/3 - t) * 6
		return p
	#

#



IntRGB.BLACK = IntRGB.parseCSS("#000000")
IntRGB.DARK_GRAY = IntRGB.parseCSS("#404040")
IntRGB.GRAY = IntRGB.parseCSS("#808080")
IntRGB.LIGHT_GRAY = IntRGB.parseCSS("#c0c0c0")
IntRGB.WHITE = IntRGB.parseCSS("#ffffff")

IntRGB.LIGHT_RED = IntRGB.parseCSS("#ff8080")
IntRGB.RED = IntRGB.parseCSS("#ff0000")
IntRGB.DARK_RED = IntRGB.parseCSS("#800000")

IntRGB.LIGHT_GREEN = IntRGB.parseCSS("#80ff80")
IntRGB.GREEN = IntRGB.parseCSS("#00ff00")
IntRGB.DARK_GREEN = IntRGB.parseCSS("#008000")

IntRGB.LIGHT_BLUE = IntRGB.parseCSS("#8080ff")
IntRGB.BLUE = IntRGB.parseCSS("#0000ff")
IntRGB.DARK_BLUE = IntRGB.parseCSS("#000080")

IntRGB.LIGHT_YELLOW = IntRGB.parseCSS("#ffff80")
IntRGB.YELLOW = IntRGB.parseCSS("#ffff00")
IntRGB.DARK_YELLOW = IntRGB.parseCSS("#808000")

IntRGB.LIGHT_VIOLET = IntRGB.parseCSS("#ff80ff")
IntRGB.VIOLET = IntRGB.parseCSS("#ff00ff")
IntRGB.DARK_VIOLET = IntRGB.parseCSS("#800080")

IntRGB.LIGHT_CYAN = IntRGB.parseCSS("#80ffff")
IntRGB.CYAN = IntRGB.parseCSS("#00ffff")
IntRGB.DARK_CYAN = IntRGB.parseCSS("#008080")

IntRGB.LIGHT_ORANGE = IntRGB.parseCSS("#ffc080")
IntRGB.ORANGE = IntRGB.parseCSS("#ff8000")
IntRGB.DARK_ORANGE = IntRGB.parseCSS("#804000")

IntRGB.LIGHT_BROWN = IntRGB.parseCSS("#c06000")
IntRGB.BROWN = IntRGB.parseCSS("#804000")
IntRGB.DARK_BROWN = IntRGB.parseCSS("#502800")

#IntRGB.NAVY = IntRGB.DARK_BLUE
#IntRGB.AQUA = IntRGB.CYAN
#IntRGB.TEAL = IntRGB.DARK_CYAN
#IntRGB.LIME = IntRGB.GREEN
#IntRGB.FUCHSIA = IntRGB.VIOLET
#IntRGB.PURPLE = IntRGB.DARK_VIOLET
#IntRGB.OLIVE = IntRGB.DARK_YELLOW
#IntRGB.SILVER = IntRGB.LIGHT_GRAY



